/*
 * 3D Engine 
 * main.c: 
 * Based on pikuma.com 3D software renderer in C
 * Copyright (c) 2021 986-Studio. All rights reserved.
 *
 * Created by Manoël Trapier on 01/03/2021.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>

#include <SDL.h>

#include <log.h>

#include <display.h>
#include <vector.h>
#include <mesh.h>

bool isRunning = true;
int previousFrameTime = 0;

vec3_t cameraPosition =
{
    .x = 0,
    .y = 0,
    .z = -5,
};
vec3_t cubeRotation =
{
    .x = 0,
    .y = 0,
    .z = 0,
};

triangle_t projectedTriangles[N_MESH_FACES];

double fovFactor = 640;

void setup()
{
    frameBuffer = (uint32_t *)calloc(windowWidth * windowHeight, sizeof(uint32_t));
    if (!frameBuffer)
    {
        log(TLOG_PANIC, NULL, "Memory allocation error.");
        goto exit;
    }

    frameBufferTexture = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_ARGB8888, SDL_TEXTUREACCESS_STREAMING, windowWidth,
                                           windowHeight);
    if (frameBufferTexture == NULL)
    {
        log(TLOG_PANIC, NULL, "SDL Texture creation error: %s", SDL_GetError());
        goto exit;
    }

exit:
    return;
}

void processInput()
{
    SDL_Event event;
    SDL_PollEvent(&event);

    switch (event.type)
    {
    case SDL_QUIT:isRunning = false;
        break;

    case SDL_KEYDOWN:
        if (event.key.keysym.sym == SDLK_ESCAPE)
        {
            isRunning = false;
        }
        break;

    default:
        break;
    }

}

vec2_t orthographicPointProjection(vec3_t point)
{
    vec2_t ret;

    ret.x = point.x * fovFactor;
    ret.y = point.y * fovFactor;

    return ret;
}

vec2_t perspectivePointProjection(vec3_t point)
{
    vec2_t ret;

    ret.x = (point.x/point.z) * (fovFactor/2.);
    ret.y = (point.y/point.z) * (fovFactor/2.);

    return ret;
}

void update()
{
    int i, j;

    uint32_t timeToWait = FRAME_TARGET_TIME - (SDL_GetTicks() - previousFrameTime);
    if (timeToWait <= FRAME_TARGET_TIME)
    {
        SDL_Delay(timeToWait);
    }
    previousFrameTime = SDL_GetTicks();

    cubeRotation.x += 0.01;
    cubeRotation.y += 0.01;
    cubeRotation.z += 0.01;

    for(i = 0; i < N_MESH_FACES; i++)
    {
        vec3_t vertices[3] =
        {
            meshVertices[meshFaces[i].a - 1],
            meshVertices[meshFaces[i].b - 1],
            meshVertices[meshFaces[i].c - 1],
        };

        for(j = 0; j < 3; j++)
        {
            vec3_t transformed = vertices[j];

            transformed = vec3RotateX(transformed, cubeRotation.x);
            transformed = vec3RotateY(transformed, cubeRotation.y);
            transformed = vec3RotateZ(transformed, cubeRotation.z);

            transformed.z -= cameraPosition.z;

            vec2_t projected = perspectivePointProjection(transformed);

            projected.x += (windowWidth  / 2);
            projected.y += (windowHeight / 2);

            projectedTriangles[i].points[j] = projected;
        }
    }
}

void render()
{
    int i;
    SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255);
    SDL_RenderClear(renderer);


    for(i = 0; i < N_MESH_FACES; i++)
    {
        drawRectangle(projectedTriangles[i].points[0].x,
                      projectedTriangles[i].points[0].y,4, 4, 0xFFFFFF00);
        drawRectangle(projectedTriangles[i].points[1].x,
                      projectedTriangles[i].points[1].y,4, 4, 0xFFFF00FF);
        drawRectangle(projectedTriangles[i].points[2].x,
                      projectedTriangles[i].points[2].y,4, 4, 0xFF00FFFF);
    }


    renderFrameBuffer();
    clearFrameBuffer(0xFF000000);

    SDL_RenderPresent(renderer);
}

int main(int argc, char *argv[])
{
    int param_i;
    bool fullScreen = false;
    MAX_DEBUG_LEVEL = TLOG_DEBUG;

    log(TLOG_ALWAYS, NULL, "Booting 3D Engine (version %s)!", VERSION);

    for (param_i = 1 ; (param_i < argc) && (argv[param_i][0] == '-') ; param_i++)
    {
        switch (argv[param_i][1])
        {
        default: exit(-1); /* Option not recognized */
        case 'f': fullScreen = true; break;
        case 'w': windowWidth = atoi(argv[++param_i]);
        case 'h': windowHeight = atoi(argv[++param_i]); break;
#ifdef DYNA_LOG_LEVEL
        case 'l': MAX_DEBUG_LEVEL = atoi(argv[++param_i]); break;
#endif
        case '-': goto no_more_params; /* Could use break, but this is more clear */
        }
    }
no_more_params:

    isRunning = initialiseWindow(fullScreen);

    setup();

    while (isRunning)
    {
        processInput();
        update();
        render();
    }

    destroy_window();

    return 0;
}